import asyncio
from functools import wraps

import pygame

def clamp(num, min_, max_):
    if num < min_:
        return min_
    elif num > max_:
        return max_
    return num

def point_touching_sprite(point, sprite):
    return sprite.left <= point.x <= sprite.right and sprite.bottom <= point.y <= sprite.top

def sprite_touching_sprite(a, b):
    if a.left >= b.right or a.right <= b.left or a.top <= b.bottom or a.bottom >= b.top:
        return False
    return True

def make_async(func):
    if asyncio.iscoroutinefunction(func): return func
    @wraps(func)
    async def async_func(*args, **kwargs):
        return func(*args, **kwargs)
    return async_func


import copy

import pygame

from .color import color_name_to_rgb
from .drawable_2d import Drawable2D

class Rect(Drawable2D):
    def __init__(self, game, color, x, y, z, angle, width, height, border_color, border_width):
        self._width = width
        self._height = height
        self._color = color
        self._border_color = border_color
        self._border_width = border_width
        super().__init__(game, x, y, z, angle)

    def _compute_primary_surface(self):
        surface = pygame.Surface((self._width, self._height), pygame.SRCALPHA)
        if self._border_width and self._border_color:
            surface.fill(color_name_to_rgb(self._border_color))
            pygame.draw.rect(
                surface,
                color_name_to_rgb(self._color),
                (
                    self._border_width,
                    self._border_width,
                    self._width - (self._border_width*2),
                    self._height - (self._border_width*2)
                )
            )
        else:
            surface.fill(color_name_to_rgb(self._color))
        return surface

    @property 
    def color(self):
        return self._color

    @color.setter
    def color(self, _color):
        self._color = _color
        self._primary_surface = self._compute_primary_surface()
        self._secondary_surface = self._compute_secondary_surface()

    @property 
    def border_color(self):
        return self._border_color

    @border_color.setter
    def border_color(self, _border_color):
        self._border_color = _border_color
        self._primary_surface = self._compute_primary_surface()
        self._secondary_surface = self._compute_secondary_surface()

    @property 
    def border_width(self):
        return self._border_width

    @border_width.setter
    def border_width(self, _border_width):
        self._border_width = _border_width
        self._primary_surface = self._compute_primary_surface()
        self._secondary_surface = self._compute_secondary_surface()


from collections import defaultdict

import pygame

from .utils import make_async

ANY_KEY = 'any_key'

class Keyboard():
    def __init__(self):
        self.pressed_keys = set()
        self.released_keys = set()
        self.keys_pressed_this_frame = set()
        self.keypress_callbacks = defaultdict(list)
        self.keyrelease_callbacks = defaultdict(list)

    def register_key_down_event(self, event):
        if event.key in key_map:
            self.pressed_keys.add(pygame_key_to_name(event))
            self.keys_pressed_this_frame.add(pygame_key_to_name(event))

    def register_key_up_event(self, event):
        if event.key in key_map:
            self.pressed_keys.remove(pygame_key_to_name(event))
            self.released_keys.add(pygame_key_to_name(event))

    def clear_frame_events(self):
        self.released_keys.clear()
        self.keys_pressed_this_frame.clear()

    def invoke_callbacks(self, task_runner):
        for key in self.keys_pressed_this_frame:
            for callback in self.keypress_callbacks[key]:
                task_runner.run(callback, key)
            for callback in self.keypress_callbacks[ANY_KEY]:
                task_runner.run(callback, key)

        for key in self.released_keys:
            for callback in self.keyrelease_callbacks[key]:
                task_runner.run(callback, key)
            for callback in self.keyrelease_callbacks[ANY_KEY]:
                task_runner.run(callback, key)

    def when_any_key_pressed(self, func):
        self.keypress_callbacks[ANY_KEY].append(make_async(func))
        return func

    def when_key_pressed(self, *keys):
        def decorator(func):
            for key in keys:
                self.keypress_callbacks[key].append(make_async(func))
            return func
        return decorator

    def when_any_key_released(self, func):
        self.keyrelease_callbacks[ANY_KEY].append(make_async(func))
        return func

    def when_key_released(self, *keys):
        def decorator(func):
            for key in keys:
                self.keyrelease_callbacks[key].append(make_async(func))
            return func
        return decorator

    def key_is_pressed(self, *keys):
        """
        Returns True if any of the given keys are pressed.
        Example:
        ```
        @nyan.repeat_forever
        async def do():
            if nyan.key_is_pressed('up', 'w'):
                print('up or w pressed')
        ```
        """
        for key in keys:
            if key in self.pressed_keys:
                return True
        return False

key_map = {
    pygame.K_BACKSPACE: 'backspace',
    pygame.K_TAB: 'tab',
    pygame.K_CLEAR: 'clear',
    pygame.K_RETURN: 'enter',
    pygame.K_PAUSE: 'pause',
    pygame.K_ESCAPE: 'escape',
    pygame.K_SPACE: 'space',
    pygame.K_EXCLAIM: '!',
    pygame.K_QUOTEDBL: '"',
    pygame.K_HASH: '#',
    pygame.K_DOLLAR: '$',
    pygame.K_AMPERSAND: '&',
    pygame.K_QUOTE: "'",
    pygame.K_LEFTPAREN: '(',
    pygame.K_RIGHTPAREN: ')',
    pygame.K_ASTERISK: '*',
    pygame.K_PLUS: '+',
    pygame.K_COMMA: ',',
    pygame.K_MINUS: '-',
    pygame.K_PERIOD: '.',
    pygame.K_SLASH: '/',
    pygame.K_0: '0',
    pygame.K_1: '1',
    pygame.K_2: '2',
    pygame.K_3: '3',
    pygame.K_4: '4',
    pygame.K_5: '5',
    pygame.K_6: '6',
    pygame.K_7: '7',
    pygame.K_8: '8',
    pygame.K_9: '9',
    pygame.K_COLON: ':',
    pygame.K_SEMICOLON: ';',
    pygame.K_LESS: '<',
    pygame.K_EQUALS: '=',
    pygame.K_GREATER: '>',
    pygame.K_QUESTION: '?',
    pygame.K_AT: '@',
    pygame.K_LEFTBRACKET: '[',
    pygame.K_BACKSLASH: '\\',
    pygame.K_RIGHTBRACKET: ']',
    pygame.K_CARET: '^',
    pygame.K_UNDERSCORE: '_',
    pygame.K_BACKQUOTE: '`',
    pygame.K_a: 'a',
    pygame.K_b: 'b',
    pygame.K_c: 'c',
    pygame.K_d: 'd',
    pygame.K_e: 'e',
    pygame.K_f: 'f',
    pygame.K_g: 'g',
    pygame.K_h: 'h',
    pygame.K_i: 'i',
    pygame.K_j: 'j',
    pygame.K_k: 'k',
    pygame.K_l: 'l',
    pygame.K_m: 'm',
    pygame.K_n: 'n',
    pygame.K_o: 'o',
    pygame.K_p: 'p',
    pygame.K_q: 'q',
    pygame.K_r: 'r',
    pygame.K_s: 's',
    pygame.K_t: 't',
    pygame.K_u: 'u',
    pygame.K_v: 'v',
    pygame.K_w: 'w',
    pygame.K_x: 'x',
    pygame.K_y: 'y',
    pygame.K_z: 'z',
    pygame.K_DELETE: 'delete',
    pygame.K_UP: 'up',
    pygame.K_DOWN: 'down',
    pygame.K_RIGHT: 'right',
    pygame.K_LEFT: 'left',
    pygame.K_INSERT: 'insert',
    pygame.K_HOME: 'home',
    pygame.K_END: 'end',
    pygame.K_PAGEUP: 'pageup',
    pygame.K_PAGEDOWN: 'pagedown',
    pygame.K_F1: 'F1',
    pygame.K_F2: 'F2',
    pygame.K_F3: 'F3',
    pygame.K_F4: 'F4',
    pygame.K_F5: 'F5',
    pygame.K_F6: 'F6',
    pygame.K_F7: 'F7',
    pygame.K_F8: 'F8',
    pygame.K_F9: 'F9',
    pygame.K_F10: 'F10',
    pygame.K_F11: 'F11',
    pygame.K_F12: 'F12',
    pygame.K_F13: 'F13',
    pygame.K_F14: 'F14',
    pygame.K_F15: 'F15',
    pygame.K_NUMLOCK: 'numlock',
    pygame.K_CAPSLOCK: 'capslock',
    pygame.K_SCROLLOCK: 'scrollock',
    pygame.K_RSHIFT: 'shift',
    pygame.K_LSHIFT: 'shift',
    pygame.K_RCTRL: 'control',
    pygame.K_LCTRL: 'control',
    pygame.K_RALT: 'alt',
    pygame.K_LALT: 'alt',
    pygame.K_RMETA: 'meta',
    pygame.K_LMETA: 'meta',
    pygame.K_LSUPER: 'super',
    pygame.K_RSUPER: 'super',
    pygame.K_EURO: '€',
}

def pygame_key_to_name(pygame_key_event):
    return key_map[pygame_key_event.key]


__version__ = '0.1.0'

import asyncio as _asyncio

import pygame as _pygame

from .game import Game as _Game
from .sprite_manager import SpriteManager as _SpriteManager
from .screen import Screen as _Screen
from .mouse import Mouse as _Mouse
from .keyboard import Keyboard as _Keyboard
from .custom_event import CustomEvent as _CustomEvent
from .sound import Sound as _Sound
from .music import Music as _Music
from .task_runner import TaskRunner as _TaskRunner
from .sprite import Sprite as _Sprite
from .text import Text as _Text
from .rect import Rect as _Rect
from .circle import Circle as _Circle
from .random import Random as _Random
from .timer import new_timer

_pygame.init()

_task_runner = _TaskRunner()
screen = _Screen()
mouse = _Mouse()
_keyboard = _Keyboard()
_custom_event = _CustomEvent()
_random = _Random(screen)
_sprite_manager = _SpriteManager()
music = _Music()
_game = _Game(_task_runner, _sprite_manager, screen, mouse, _keyboard, _custom_event)

when_mouse_clicked = mouse.when_clicked
when_mouse_click_released = mouse.when_click_released

key_is_pressed = _keyboard.key_is_pressed
when_key_pressed = _keyboard.when_key_pressed
when_key_released = _keyboard.when_key_released
when_any_key_pressed = _keyboard.when_any_key_pressed
when_any_key_released = _keyboard.when_any_key_released

broadcast = _custom_event.broadcast
when_event_recieved = _custom_event.when_event_recieved

random_color = _random.random_color
random_number = _random.random_number
random_position = _random.random_position

get_sprites = _sprite_manager.get_sprites

def new_sprite(image, x=0, y=0, z=0, angle=0):
    return _Sprite(_sprite_manager, image, x, y, z, angle)

def new_text(text, x=0, y=0, z=0, angle=0, font=None, font_size=50, color='black'):
    return _Text(_sprite_manager, text, x, y, z, angle, font, font_size, color)

def new_rect(color='black', x=0, y=0, z=0, angle=0, width=100, height=200, border_color='light blue', border_width=0):
    return _Rect(_sprite_manager, color, x, y, z, angle, width, height, border_color, border_width)

def new_circle(color='black', x=0, y=0, z=0, angle=0, radius=100, border_color='light blue', border_width=0):
    return _Circle(_sprite_manager, color, x, y, z, angle, radius, border_color, border_width)

def new_sound(sound):
    return _Sound(sound)

def when_program_starts(func):
    """
    Call code right when the program starts.
    Used like this:
    ```
    @nyan.when_program_starts
    def do():
        print('the program just started!')
    ```
    """
    _game.register_when_program_starts_callbacks(func)
    return func

def repeat_forever(func):
    """
    Calls the given function repeatedly in the game loop.
    Example:
    ```
    text = nyan.new_text('hello world')
    @nyan.repeat_forever
    async def do():
        text.turn(degrees=15)
    ```
    """
    _game.register_forever_callback(func)
    return func

"""
TODO: determine the best name for this function. these are the current
candidates, other names are also welcome:

- @foreach_sprite_with_tag('player', 'bullet')
- @foreach_sprite('player', 'player')
- @foreach('player', 'bullet')

I am leaning toward the first one, but the name is somewhat confusing
when multiple tags are passed. for example, will this function
`@foreach_sprite_with_tag('player', 'bullet')` return sprites that has
both tags or one of them?
"""
def foreach_sprite(*tags):
    """
    Calls the given function for each sprite that has any of the passed tags.
    To be used in conjunction with `@repeat_forever` and `@when_program_starts`
    decorators. Example:
    ```
    @nyan.repeat_forever
    @nyan.foreach_sprite('projectile')
    async def propel_projectile(projectile):
        projectile.move(25)
    ```
    """
    def decorator(func):
        func.tags = tags
        return func
    return decorator

async def sleep(seconds=1.0):
    """
    Wait a number of seconds. Used with the await keyword like this:
    ```
    @nyan.repeat_forever
    async def do():
        await nyan.sleep(seconds=2)
        print('hi')
    ```
    """
    await _asyncio.sleep(seconds)

def start_program():
    """
    Calling this function starts your program running.
    `nyan.start_program()` should almost certainly go at the very end of your program.
    """
    _game.invoke_when_program_starts_callbacks()
    _task_runner.loop.call_soon(_game.run)
    try:
        _task_runner.loop.run_forever()
    finally:
        _pygame.quit()


import os

import pygame

from .drawable_2d import Drawable2D

class Sprite(Drawable2D):
    def __init__(self, game, image, x, y, z, angle):
        self._image = image
        super().__init__(game, x, y, z, angle)

    def _compute_primary_surface(self):
        return pygame.image.load(os.path.join('assets', self._image)).convert_alpha()
 
    @property
    def image(self):
        return image

    @image.setter
    def image(self, image):
        self._image = image
        self._primary_surface = self._compute_primary_surface()
        self._secondary_surface = self._compute_secondary_surface()


from collections import defaultdict

class SpriteManager():
    def __init__(self):
        self.all_sprites = []
        self.sprite_groups = defaultdict(set)

    def get_sprites(self, *tags):
        if len(tags) == 0:
            return self.all_sprites
        elif len(tags) == 1:
            return self.sprite_groups[tags[0]]
        else:
            sprites = set()
            for tag in tags:
                sprites |= self.sprite_groups[tag]
            return sprites

    def add_sprite_to_group(self, sprite, group):
        self.sprite_groups[group].add(sprite)

    def remove_sprite_from_group(self, sprite, group):
        self.sprite_groups[group].remove(sprite)

    def register_sprite(self, sprite):
        self.all_sprites.append(sprite)
        for tag in sprite._tags:
            self.add_sprite_to_group(sprite, tag)

    def unregister_sprite(self, sprite):
        self.all_sprites.remove(sprite)
        for tag in sprite._tags:
            self.remove_sprite_from_group(sprite, tag)

    def invoke_callbacks(self, task_runner, mouse):
        if not mouse._is_clicked_this_frame: return

        for sprite in self.all_sprites:
            if sprite.is_touching(mouse):
                sprite._invoke_when_clicked_callbacks(task_runner)

    def draw(self, screen):
        self.all_sprites = sorted(self.all_sprites, key=lambda sprite: sprite.z)
        for sprite in self.all_sprites:
            sprite._draw(screen)



import math
from abc import ABC, abstractmethod
from copy import copy

import pygame

from .utils import clamp, sprite_touching_sprite, point_touching_sprite, make_async

class Drawable2D(ABC):
    def __init__(self, game, x, y, z, angle):
        self._game = game
        self.x = x
        self.y = y
        self.z = z
        self._angle = angle

        self._size = 100
        self._transparency = 100
        self._brightness = 0

        self.is_hidden = False
        self._tags = set()
        self._when_clicked_callbacks = []

        self._primary_surface = self._compute_primary_surface()
        self._secondary_surface = self._compute_secondary_surface()
        self._game.register_sprite(self)

    @abstractmethod
    def _compute_primary_surface(self):
        pass

    def _compute_secondary_surface(self):
        surface = pygame.transform.rotate(self._primary_surface, self._angle)
        surface = pygame.transform.scale(
            surface,
            (round(surface.get_width() * (self._size/100)),
            round(surface.get_height() * (self._size/100)))
        )
        scaled_transparency = round((self._transparency/100) * 255)
        surface.fill((255, 255, 255, scaled_transparency), special_flags=pygame.BLEND_RGBA_MULT)

        scaled_brightness = round((abs(self._brightness)/100) * 255)
        brightness_blend_mode = pygame.BLEND_RGB_ADD if self._brightness > 0 else pygame.BLEND_RGB_SUB
        surface.fill((scaled_brightness, scaled_brightness, scaled_brightness), special_flags=brightness_blend_mode)

        return surface

    def _invoke_when_clicked_callbacks(self, task_runner):
        for callback in self._when_clicked_callbacks:
            task_runner.run(callback)

    def when_clicked(self, callback):
        self._when_clicked_callbacks.append(make_async(callback))

    def add_tag(self, tag):
        self._tags.add(tag)
        self._game.add_sprite_to_group(self, tag)
 
    def remove_tag(self, tag):
        self._tags.remove(tag)
        self._game.remove_sprite_from_group(self, tag)

    def clone(self):
        clone = copy(self)
        clone._tags = clone._tags.copy()
        self._game.register_sprite(clone)
        return clone

    def remove(self):
        self._game.unregister_sprite(self)

    @property 
    def angle(self):
        return self._angle

    @angle.setter
    def angle(self, angle):
        self._angle = angle
        self._secondary_surface = self._compute_secondary_surface()

    @property 
    def size(self):
        return self._size

    @size.setter
    def size(self, size):
        self._size = clamp(size, 0, math.inf)
        self._secondary_surface = self._compute_secondary_surface()

    @property
    def transparency(self):
        return self._transparency

    @transparency.setter
    def transparency(self, amount):
        self._transparency = clamp(amount, 0, 100)
        self._secondary_surface = self._compute_secondary_surface()

    @property
    def brightness(self):
        return self._brightness

    @brightness.setter
    def brightness(self, value):
        self._brightness = clamp(value, -100, 100)
        self._secondary_surface = self._compute_secondary_surface()

    def move(self, steps):
        angle = math.radians(self._angle)
        self.x += steps * math.cos(angle)
        self.y += steps * math.sin(angle)

    def turn(self, degrees):
        self._angle = (self._angle + degrees) % 360
        self._secondary_surface = self._compute_secondary_surface()

    def hide(self):
        self.is_hidden = True

    def show(self):
        self.is_hidden = False

    @property
    def is_shown(self):
        return not self._is_hidden

    @is_shown.setter
    def is_shown(self, show):
        self._is_hidden = not show

    def is_touching(self, sprite_or_point):
        if isinstance(sprite_or_point, Drawable2D):
            return sprite_touching_sprite(sprite_or_point, self)
        else:
            return point_touching_sprite(sprite_or_point, self)

    def point_towards(self, x, y=None):
        try:
            # x can either be a number or a sprite. If it's a sprite:
            x, y = x.x, x.y
        except AttributeError:
            x, y = x, y
        self._angle = math.degrees(math.atan2(y - self.y, x - self.x))

    def go_to(self, x, y=None):
        try:
            # x can either be a number or a sprite. If it's a sprite:
            self.x = x.x
            self.y = x.y
        except AttributeError:
            self.x = x
            self.y = y

    def distance_to(self, x, y=None):
        try:
            # x can either be a number or a sprite. If it's a sprite:
            x1 = x.x
            y1 = x.y
        except AttributeError:
            x1 = x
            y1 = y

        dx = self.x - x1
        dy = self.y - y1

        return math.sqrt(dx**2 + dy**2)

    @property
    def width(self):
        return self._secondary_surface.get_width()

    @property
    def height(self):
        return self._secondary_surface.get_height()

    @property
    def right(self):
        return self.x + self.width / 2

    @right.setter
    def right(self, x):
        self.x = x - self.width / 2

    @property
    def left(self):
        return self.x - self.width / 2

    @left.setter
    def left(self, x):
        self.x = x + self.width / 2

    @property
    def top(self):
        return self.y + self.height / 2

    @top.setter
    def top(self, y):
        self.y = y - self.height / 2

    @property
    def bottom(self):
        return self.y - self.height / 2

    @bottom.setter
    def bottom(self, y):
        self.y = y + self.height / 2

    # replace screen with camera object
    def _draw(self, screen):
        if self.is_hidden: return
        x = self.x + (screen.width/2) - (self.width/2)
        y = (screen.height/2) - self.y - (self.height/2)
        screen._surface.blit(self._secondary_surface, (x, y))


import pygame

class Screen():
    def __init__(self, width=800, height=600):
        self._width = width
        self._height = height
        self._surface = pygame.display.set_mode((self._width, self._height))

    @property
    def width(self):
        return self._width

    @width.setter
    def width(self, width):
        self._width = width
        self._surface = pygame.display.set_mode((self._width, self._height))

    @property
    def height(self):
        return self._height

    @height.setter
    def height(self, height):
        self._height = height
        self._surface = pygame.display.set_mode((self._width, self._height))

    @property
    def top(self):
        return self.height / 2

    @property
    def bottom(self):
        return self.height / -2

    @property
    def left(self):
        return self.width / -2

    @property
    def right(self):
        return self.width / 2


import asyncio

class TaskRunner():
    def __init__(self):
        self.loop = asyncio.get_event_loop()
        self.running_tasks = set()

    def run(self, task, *args):
        if (task, *args) in self.running_tasks:
            return False

        self.running_tasks.add((task, *args))
        def on_task_end(f): self.running_tasks.remove((task, *args))
        future = self.loop.create_task(task(*args))
        future.add_done_callback(on_task_end)
        return True
